#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include "gensock.h"

#ifdef WIN32
  #define __far far
  #define huge far
  #define __near near
#endif

#define MAXOUTLINE 65536

typedef int (FAR PASCAL *proto_pgensock_connect) (char FAR * hostname, char FAR * service, socktag FAR * pst, void * tid);
typedef int (FAR PASCAL *proto_pgensock_getchar) (socktag st, int wait, char FAR * ch);
typedef int (FAR PASCAL *proto_pgensock_put_data) (socktag st, char FAR * data, unsigned long length);
typedef int (FAR PASCAL *proto_pgensock_close) (socktag st);
typedef int (FAR PASCAL *proto_pgensock_gethostname) (char FAR * name, int namelen);
typedef int (FAR PASCAL *proto_pgensock_put_data_buffered) (socktag st, char FAR * data, unsigned long length);
typedef int (FAR PASCAL *proto_pgensock_put_data_flush) (socktag st);

proto_pgensock_connect pgensock_connect;
proto_pgensock_getchar pgensock_getchar;
proto_pgensock_put_data pgensock_put_data;
proto_pgensock_close pgensock_close;
proto_pgensock_gethostname pgensock_gethostname;
proto_pgensock_put_data_buffered pgensock_put_data_buffered;
proto_pgensock_put_data_flush pgensock_put_data_flush;

// Buffer overflows just waiting to happen.

#define SERVER_SIZE        256
#define SENDER_SIZE        256

#define BYTENUMRETRIES     256
#define CONNECTNUMRETRIES  4

//static HANDLE gensock_lib = 0;
//static char my_hostname[1024];

typedef struct {
       socktag FNGRSock;
       char FNGRProxy[SERVER_SIZE];
       char FNGRHost[SERVER_SIZE];
       char Sender[SENDER_SIZE];
       int port;
       int proxyport;
       int quiet;
       int content_length;
       int bytes_read;
       int BinaryMode;
       int skip_bytes;
       FILE * output_file;
       char * output_filename;
       int TypeSpecified;
       int cycle_count;
       char * cycle_url;
       int DEBUG;
} FNGSTRUCT;

int multithreaded=0;

#define NMTH(x) if(!multithreaded) x;

#if 0
#ifdef _DEBUG
void complain (char * message)
{
  OutputDebugString (message);
}
#else
void complain (char * message)
{
  MessageBox (NULL, message, "GENSOCK.DLL Error", MB_OK|MB_ICONHAND);
}
#endif
#else

void complain (char * message)
{
  fprintf( stderr, "%s\n",message );
}

#endif

void gensock_error (char * function, int retval) {

  complain(function);

  switch( retval ) {

   /* Most definately errors */

   case 4001: complain( "Error: Malloc failed (possibly out of memory)."); break;
   case 4002: complain( "Error: Error sending data."); break;
   case 4003: complain( "Error: Error initializing winsock."); break;
   case 4004: complain( "Error: Version not supported."); break;
   case 4005: complain( "Error: The winsock version specified by gensock is not supported by this winsock program."); break;

   /* Probably error -- not connected */

   case 4006: complain( "Error: Network not ready."); break;
   case 4007: complain( "Error: Can't resolve (finger server) hostname."); break;
   case 4008: complain( "Error: Can't create a socket (too many simultaneous links?)"); break;
   case 4009: complain( "Error: Error reading socket."); break;
   case 4010: complain( "Error: Not a socket."); break;

   /* Possibly recoverable error */

   case 4011: complain( "Error: Busy."); break;
   case 4012: complain( "Error: Error closing socket."); break;
   case 4013: complain( "Error: Wait a bit (possible timeout)."); break;
   case 4014: complain( "Error: Can't resolve service."); break;
   case 4015: complain( "Error: Can't connect to webserver (timed out if winsock.dll error 10060)"); break;

   /* Legitimate termination condition */

   case 4016: complain( "Error: Connection to webserver was dropped."); break;
   case 4017: complain( "Error: Server refused connection."); break;
   default:   fprintf( stderr, "error %d in function '%s'\n", retval, function ); break;
  }
}

#define errormsg(msg) {                        \
 if( !quiet ) fprintf(stderr,"%s\n",msg);      \
 return -1;                                    \
}

#define assert(expr,msg) { if( !expr ) errormsg(msg); }

// Attach straight to gensock module.  (This used to be a DLL, GetProcAddress
// hook up thingy.)

#define GetAddr(name) {   \
  pgensock_##name = (proto_pgensock_##name) gensock_##name;	\
}

int load_gensock(void) {

  GetAddr(connect);
  GetAddr(getchar);
  GetAddr(put_data);
  GetAddr(close);
  GetAddr(gethostname);
  GetAddr(put_data_buffered);
  GetAddr(put_data_flush);

  return 0;
}

int open_finger_socket( FNGSTRUCT * ls ) {
int retry,retval,l;
char c;
char * host;
int port;

  /* load the library if it's not loaded */

  retval = load_gensock();
  if( retval ) return retval;

  if( ls->FNGRProxy[0] != '\0' ) {

    host = ls->FNGRProxy;
    port = ls->proxyport;

  } else {

    host = ls->FNGRHost;
    port = ls->port;

  }

  // Has the port been explicitely specified?  (As in hostname:port)

  for(l=0;host[l]!=':' && host[l]!='\0';l++) {}

  if( host[l]==':' ) {
      c = host[l];
      host[l] = '\0';

      // QUIET fprintf(stderr,"{Connecting: %s port %s}\n",host,(host+l+1));

      retry = CONNECTNUMRETRIES;
      goto retrystart0;
      do {
          fprintf(stderr,"{Retrying: %s:%s}\n",host,(host+l+1));
          Sleep(0);
          retrystart0:;
          retval = (*pgensock_connect) ((LPSTR) host,(LPSTR)(host+l+1),&(ls->FNGRSock),ls);
          retry--;
      } while( (retval==4015) && retry>0 );

      if( retval!=0 ) {

          if( retval==ERR_CONNECTION_REFUSED ) {
              fprintf(stderr,"{Connection refused from host %s}\n",host);
          } else {
              fprintf(stderr,"{Gensock error:%d from host %s}\n",retval,host);
          }
      }

      ls->FNGRHost[l] = c;
  } else {
      retval = (*pgensock_connect) ((LPSTR) host,(LPSTR)"http",&(ls->FNGRSock),ls);
      if( retval == ERR_CANT_RESOLVE_SERVICE ) {
          char ports[256];
          sprintf( ports, "%d", port );

          retry = CONNECTNUMRETRIES;
          goto retrystart1;
          do {
              fprintf(stderr,"{Retrying: %s}\n",host);
              Sleep(0);
              retrystart1:;
              retval = (*pgensock_connect) ((LPSTR)host,(LPSTR)ports,&(ls->FNGRSock),ls);
              retry--;
          } while( (retval==4015) && retry>0 );

          if( retval!=0 ) {
              if( retval==ERR_CONNECTION_REFUSED ) {
                  fprintf(stderr,"{Connection refused from host %s}\n",host);
              } else {
                  fprintf(stderr,"{Gensock error:%d from host %s}\n",retval,host);
              }
          }

          //retval = (*pgensock_connect) ((LPSTR)host,(LPSTR)ports,&(ls->FNGRSock),ls);
      }
  }

  // retval = (*pgensock_gethostname) (my_hostname, sizeof(my_hostname));

  if( retval==0 ) return 0;

  if( ls->quiet ) gensock_error ("gensock_connect", retval);
  return -1;
}

#define GSCALL( name, parms ) {                 \
  retval = (*pgensock_##name)parms;             \
  if( retval ) {                                \
    gensock_error ("gensock_"#name, retval);    \
    return -1;                                  \
  }                                             \
}

#define strGSCALL( name, parms ) {              \
  retval = (*pgensock_##name)parms;             \
  if( retval!=NULL ) {                          \
    gensock_error ("gensock_"#name, retval);    \
    return NULL;                                \
  }                                             \
}

int close_finger_socket( FNGSTRUCT * ls ) {
int retval;

  retval = (*pgensock_close)(ls->FNGRSock);
  if( retval!=0 && retval!=4016 ) {
    gensock_error ("gensock_close", retval);
  } else retval = 0;

  return retval;
}

char * get_finger_line( FNGSTRUCT * ls ) {
char ch;
static char in_data [MAXOUTLINE];
char * index;
int retval;
int retry;

    ch = '\0';           /* Probably not necessary */
    index = in_data;

    do {

      retry = BYTENUMRETRIES;
      goto retrystart2;
      do {
          Sleep(0);
          retrystart2:;
          retval = (*pgensock_getchar)(ls->FNGRSock, 0, &ch);
          retry--;
      } while( retval>=4011 && retval<=4015 && retry>0 );

      if( retval!=0 ) {
        if( retval!=4016 ) gensock_error ("gensock_getchar", retval);
        return NULL;
      }

      *index = ch;
      ls->content_length--;
      ls->bytes_read++;

      if( ch!=13 && ch!=10 ) index++;
    } while( ch!='\0' && ch!=10 );

    index[0] = '\0';

    return in_data;
}

int put_finger_line( socktag sock, char far * line, unsigned int nchars ) {
int retval;

    GSCALL(put_data,(sock, line, (unsigned long) nchars));
    return 0;
}

int putline_internal (socktag sock, char * line, unsigned int nchars) {
int retval;

    retval = (*pgensock_put_data)(sock,(char FAR *)line,(unsigned long) nchars);
    if( retval==0 ) return 0;

    switch (retval) {
     case ERR_NOT_CONNECTED:
      gensock_error( "FNGR server has closed the connection", retval );
      break;

     default:
      gensock_error ("gensock_put_data", retval);
    }
    return -1;
}

// create a registry entries for this program
int CreateRegEntry( FNGSTRUCT * ls ) {
  HKEY   hKey1;
  DWORD  dwDisposition;
  LONG   lRetCode;

  /* try to create the .INI file key */
  lRetCode = RegCreateKeyEx ( HKEY_LOCAL_MACHINE,
                              "SOFTWARE\\Public Domain\\finger",
                              0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE,NULL, &hKey1,&dwDisposition
                            );

  /* if we failed, note it, and leave */
  if (lRetCode != ERROR_SUCCESS) {
    if( !ls->quiet ) fprintf ( stderr, "Error in creating finger key in the registry\n");
    return 10;
  }

  /* try to set a section value */
  lRetCode = RegSetValueEx( hKey1,"FNGR client",0,REG_SZ, (BYTE *) &ls->FNGRHost[0], (strlen(ls->FNGRHost)+1));

  /* if we failed, note it, and leave */
  if (lRetCode != ERROR_SUCCESS) {
    if( ! ls->quiet ) fprintf ( stderr, "Error in setting FNGR client value in the registry\n");
    return 11;
  }

  /* try to set another section value */
  lRetCode = RegSetValueEx( hKey1,"Sender",0,REG_SZ, (BYTE *) &ls->Sender[0], (strlen(ls->Sender)+1));

  /* if we failed, note it, and leave */
  if (lRetCode != ERROR_SUCCESS) {
   if( ! ls->quiet ) fprintf ( stderr, "Error in setting sender address value in the registry\n");
    return 11;
  }

  return 0;
}

// get the registry entries for this program
int GetRegEntry( FNGSTRUCT * ls )
{
  HKEY   hKey1;
  DWORD  dwType;
  DWORD  dwBytesRead;
  LONG   lRetCode;

  // open the registry key in read mode
  lRetCode = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
                           "SOFTWARE\\Public Domain\\finger",
                           0, KEY_READ, &hKey1
                         );
  if( lRetCode != ERROR_SUCCESS ) {
     if( ! ls->quiet ) fprintf( stderr, "Failed to open registry key for WebSnarf\n" );
     return 12;
  }
  // set the size of the buffer to contain the data returned from the registry
  // thanks to Beverly Brown "beverly@datacube.com" and "chick@cyberspace.com" 
  // for spotting it...
  dwBytesRead=SERVER_SIZE;
  // read the value of the FNGR client entry
  lRetCode = RegQueryValueEx( hKey1, "FNGR client", NULL , &dwType, (BYTE *) ls->FNGRHost, &dwBytesRead);
  // if we failed, note it, and leave
  if( lRetCode != ERROR_SUCCESS ) {
    if( ! ls->quiet ) fprintf( stderr, "Failed to read FNGR client value from the registry\n" );
    return 12;
  }

  dwBytesRead=SENDER_SIZE;
  // read the value of the FNGR client entry
  lRetCode = RegQueryValueEx( hKey1, "Sender", NULL , &dwType, (BYTE *) ls->Sender, &dwBytesRead);
  // if we failed, note it, and leave
  if( lRetCode != ERROR_SUCCESS ) {
    if( !ls->quiet ) fprintf( stderr, "Failed to read senders user name from the registry\n" );
    return 12;
  }
  return 0;
}

int GetURL(FNGSTRUCT * ls, char *s);

enum { DISPLAY, HEADER, HIDE };

static int fetch(FNGSTRUCT * ls, char *server, char *url, char * tgt_fn) {
  char *s;
  int retcode;
  int retry;
  int glob_retry = CONNECTNUMRETRIES;
  char CMDline[SERVER_SIZE];
  int InitialOffset=0;
  int Continuations=0;
  int InitialTry = -1;

  ls->content_length = -1;

  if( tgt_fn!=NULL ) ls->output_file = fopen( tgt_fn, "w" );

  sprintf(CMDline,"%s\r\n",url);

// Attach to socket.

  strcpy(ls->FNGRHost,server);
  retcode = open_finger_socket( ls ) ||                            \
            put_finger_line( ls->FNGRSock, CMDline, strlen(CMDline) );

  if( retcode==0 ) {
    do {
        s = get_finger_line(ls);
        if( s==NULL ) {
            retry = CONNECTNUMRETRIES;
            do {

                // retry to make sure that this is a real termination
                // condition.

                s = get_finger_line(ls);
                retry--;
            } while( s==NULL && retry>0 );
            if( s==NULL ) break;  // Termination condition.
        }
        fprintf( ls->output_file, "%s\n", s );
    } while(1);

    close_finger_socket( ls );
  }

  if( ls->output_file!=stdout ) fclose( ls->output_file );

  return retcode;
}

int GetURL(FNGSTRUCT * ws, char *s) {
int l;

     for(l=0;s[l]!='@';l++) { 
         if( s[l]=='\0' ) return -1; 
     }

     s[l]='\0';
     fetch(ws,s+l+1,s,ws->output_filename);

     return 0;
}

#include <process.h>

int gmain(FNGSTRUCT * ws, int argc, char *argv[]);

typedef struct {
	int argc;
	char ** argv;
	FNGSTRUCT * ws;
} mainargs_type;

static volatile int NUM_THREADS = 0;

// This is atomic -- its one instruction meaning it can modify a shared
// variable with impunity.

void AtomicAdd(int * a, int b);
#pragma aux AtomicAdd =			\
"	add	[eax],ebx	"	\
parm [eax] [ebx];

static void MainThread(PVOID ma) {

    if( ma!= NULL ) {
        gmain( ((mainargs_type *)ma)->ws, ((mainargs_type *)ma)->argc, ((mainargs_type *)ma)->argv );
        free(ma);
    }

    AtomicAdd((int *)&NUM_THREADS,-1);
    _endthread ();
}

// typedef void (FAR PASCAL * proto_sleep)(unsigned int);

//static void FAR PASCAL bogus_sleep(unsigned int x) {
//    while( x>0 ) {
//        AtomicAdd((int *)&x,-1);
//    }
//}

//proto_sleep sleep = bogus_sleep;

//void FAR PASCAL Sleep(unsigned int t) {
//    sleep(t);
//}

//void INT3(void);
//#pragma aux INT3 = "int 3";

int gmain(FNGSTRUCT * ls, int argc, char *argv[]) {
int i, r=0;
mainargs_type * ma;

  if( argc<2 ) {
    return 0;
  }

  if( argc==2 && argv[1][0] == '@' ) {
    FILE * fp;
    char *s;
    char **ss;
    int n, c;

    fp = fopen( argv[1]+1, "r" );
    if( fp==NULL ) {
      fprintf ( stderr, "Error, could not open %s\n",argv[1]+1 );
      return -1;
    }

    // Buffer overflow.

    //fprintf(stderr, "{%s}\n",argv[1]);
    s = (char *)malloc(sizeof(char)*1024);

    while( NULL != fgets( s, 1023, fp ) ) {

      // Buffer overflow.

      ss = (char **)malloc(sizeof(char *)*32);

      // Decompose each line into (argc, argv) argument structure.

      ss[0] = argv[0];
      for(n=1,r=0,c=0;s[c]!=0;c++) {
        if( r==0 ) {
          ss[n] = s+c;
          n++;
          r = 1;
        }
        if(s[c]<=' ') {
          s[c] = '\0';
          do {
              c++;
          } while(s[c]<=' ' && s[c]!=0);
          c--;
          r = 0;
        }
      }

      ma = (mainargs_type *)malloc(sizeof(mainargs_type));
      if( ma==NULL ) {
        complain("malloc failed -- probably out of memory.");
        return -1;
      }

      if( n>0 ) {

          ma->argc = n;
          ma->argv = ss;
          ma->ws = (FNGSTRUCT *)malloc(sizeof(FNGSTRUCT));

          AtomicAdd((int *)&NUM_THREADS,1);

          multithreaded = 1;

          _beginthread(MainThread, 0, ma);

//_WCRTLINK extern unsigned long _beginthread( 
//                register void (*__start_address)(void *),
//                unsigned __stack_size, void *__arglist );

          while( NUM_THREADS>16 ) {
              // Surrender time slice.
              Sleep(0);
          }
      }

      s = (char *)malloc(sizeof(char)*1024);

    }

    free(s);

    fclose( fp );

    // Just sit around and twiddle our thumbs until we are done.

    while( NUM_THREADS>0 ) {
        // Surrender time slice.
        Sleep(0);
    }

    return r;
  }

  ls->quiet=0;
  ls->FNGRProxy[0] = '\0';
  ls->BinaryMode = 1;
  ls->output_file = stdout;
  ls->output_filename = NULL;
  ls->TypeSpecified = 0;
  ls->bytes_read=0;
  ls->skip_bytes = 0;
  ls->cycle_count=1;
  ls->cycle_url=NULL;
  ls->port=79;
  ls->proxyport=74;
  ls->DEBUG = 0;

  for(i=1;;i++) {
    if( argv[i][0] == '-' || argv[i][0] =='/' ) {
      switch( argv[i][1] ) {
	case 'o': case 'O':
          ls->output_filename = &(argv[i][2]);
          break;
        case 'p': case 'P':
          sscanf(&(argv[i][2]),"%d",&ls->port);
          break;
	case '?':
          fprintf(stderr,"%s [-[oFilename|pPort]] URL\n%s @[argument-filename]\n\nOptions:\n -oFilename\tOutput to file\n -pPort\t\tSet default finger port\n\n",argv[0],argv[0]);
          return 0;
        default:
          fprintf(stderr,"Bad option");
          return -1;
      }
    } else break;
  }

  GetURL(ls,argv[i]);
  return 0;
}

int main(int argc, char *argv[]) {
FNGSTRUCT main_ls;

    return gmain(&main_ls,argc,argv);
}

